import logging
from datetime import datetime
from typing import List

from bs4 import BeautifulSoup

from .data_source.base import SourceDocument


class DocumentChunker:
    def __init__(self, text_splitter, minimal_chunk_size: int = 128, add_metadata: bool = False):
        self.text_splitter = text_splitter
        self.minimal_chunk_size = minimal_chunk_size
        self.add_metadata = add_metadata

    def to_chunks(self, source_document: SourceDocument) -> List[dict]:
        logging.getLogger('statbox').info({
            'action': 'chunking',
            'document_id': source_document.document_id,
            'mode': 'cybrex',
        })
        document = source_document.document
        abstract = document.get('abstract', '')
        content = document.get('content')
        if not content:
            return []

        abstract_soup = BeautifulSoup(abstract, features='lxml')
        content_soup = BeautifulSoup(content, features='lxml')

        extracted_texts = []
        for section_id, section in enumerate(list(abstract_soup.children) + list(content_soup.children)):
            for el in list(section.find_all()):
                if el.name in {'ref', 'formula', 'math', 'figure'}:
                    el.extract()
            text = section.get_text(' ', strip=True)
            if len(text) < self.minimal_chunk_size:
                continue
            extracted_texts.append(text)

        chunks = []

        for chunk_id, chunk in enumerate(self.text_splitter.split_text('\n\n'.join(extracted_texts))):
            if len(chunk) < self.minimal_chunk_size:
                continue
            parts = []
            if self.add_metadata:
                if 'title' in document:
                    parts.append(f'TITLE: {document["title"]}')
                if 'issued_at' in document:
                    issued_at = datetime.utcfromtimestamp(document['issued_at'])
                    parts.append(f'YEAR: {issued_at.year}')
                if 'metadata' in document and 'keywords' in document['metadata']:
                    keywords = ', '.join(document['metadata']['keywords'])
                    parts.append(f'KEYWORDS: {keywords}')
                if 'tags' in document:
                    tags = ', '.join(document['tags'])
                    parts.append(f'TAGS: {tags}')
            parts.append(chunk)
            text = '\n'.join(parts)
            chunks.append({
                'real_text': text,
                'text': chunk,
                'document_id': source_document.document_id,
                'length': len(text),
                'chunk_id': chunk_id,
                'title': document["title"],
            })
        return chunks



import uuid
from typing import List

from qdrant_client import QdrantClient
from qdrant_client.models import Distance, Filter, FieldCondition, MatchValue, PointStruct, Range, VectorParams

from .base import BaseVectorStorage


class QdrantVectorStorage(BaseVectorStorage):
    def __init__(self, qdrant_config, collection_name, embedding_function, force_recreate: bool = False):
        self.db = QdrantClient(**qdrant_config)
        self.collection_name = collection_name
        self.embedding_function = embedding_function
        self.force_recreate = force_recreate

    def _ensure_collection(self, collection_name, size):
        if not self.force_recreate:
            collections = self.db.get_collections()
            for collection in collections.collections:
                if collection.name == collection_name:
                    return
            self.db.create_collection(
                collection_name=collection_name,
                vectors_config=VectorParams(size=size, distance=Distance.COSINE),
            )
        else:
            self.db.recreate_collection(
                collection_name=collection_name,
                vectors_config=VectorParams(size=size, distance=Distance.COSINE),
            )
            self.force_recreate = False
        self.db.create_payload_index(
            collection_name=collection_name,
            field_name="document_id",
            field_schema="keyword",
        )

    def get_by_field_value(self, field, value, sort: bool = True) -> List[dict]:
        points, _ = self.db.scroll(
            collection_name=self.collection_name,
            scroll_filter=Filter(
                should=[
                    FieldCondition(
                        key=field,
                        match=MatchValue(value=value)
                    ),
                ],
            ),
            limit=2 ** 31,
        )
        payloads = [point.payload for point in points]
        if sort:
            return list(sorted(payloads, key=lambda x: x['chunk_id']))
        return payloads

    def query(self, query_embedding, n_chunks: int, where: dict = None):
        self._ensure_collection(
            collection_name=self.collection_name,
            size=len(query_embedding),
        )
        query_filter = None
        if where:
            query_filter = Filter(
                must=[
                    FieldCondition(
                        key=k,  # Condition based on values of `rand_number` field.
                        range=Range(
                            gte=v,
                            lte=v
                        )
                    ) for k, v in where.items()
                ]
            )
        points = self.db.search(
            collection_name=self.collection_name,
            query_vector=query_embedding,
            query_filter=query_filter,
            limit=n_chunks,
        )
        return [{**point.payload, 'score': point.score} for point in points]

    def upsert(self, chunks: List[dict]):
        if not chunks:
            return
        if 'embedding' in chunks[0]:
            embeddings = [chunk.pop('embedding') for chunk in chunks]
        else:
            embeddings = self.embedding_function([
                chunk.pop('real_text', None) or chunk['text']
                for chunk in chunks
            ])
        self._ensure_collection(
            collection_name=self.collection_name,
            size=len(embeddings[0]),
        )
        return self.db.upsert(
            collection_name=self.collection_name,
            points=[
                PointStruct(
                    id=str(uuid.uuid1()),
                    vector=embedding,
                    payload=chunk
                )
                for chunk, embedding in zip(chunks, embeddings)
            ]
        )



from dataclasses import dataclass
from typing import (
    List,
    Optional,
)


@dataclass
class SourceDocument:
    document: dict
    document_id: str


class BaseDataSource:
    async def query_documents(self, query: str, limit: int = 5, sources: Optional[List[str]] = None) -> List[SourceDocument]:
        raise NotImplementedError()



import json
import logging
import re
import sys
import textwrap
from typing import Optional

import fire
from termcolor import colored

from .cybrex_ai import CybrexAI


class CybrexCli:
    def __init__(self, cybrex: Optional[CybrexAI] = None):
        self.cybrex = cybrex or CybrexAI()

    async def add_all_documents(self):
        async with self.cybrex as cybrex:
            async for document in cybrex.geck.get_summa_client().documents('nexus_science'):
                document = json.loads(document)
                await self.cybrex.upsert_documents([document])

    async def export_chunks(self, query: str, output_path: str, n_documents: int = 10):
        """
        Store STC text chunks in ZIP archive

        :param query: query to STC
        :param output_path: where to store result
        :param n_documents: the number of chunks to extract
        """
        async with self.cybrex as cybrex:
            print(f"{colored('Q', 'green')}: {query}")
            await cybrex.export_chunks(
                query=query,
                output_path=output_path,
                n_documents=n_documents
            )

    async def chat_doc(self, document_query: str, query: str, n_chunks: int = 5, minimum_score: float = 0.5):
        """
        Ask a question about content of document identified by DOI.

        :param document_query: query that returns unique document
        :param query: Text query to the document
        :param n_chunks: the number of chunks to extract
            more means more tokens to use and more precision in answer
        """
        async with self.cybrex as cybrex:
            print(f"{colored('Document', 'green')}: {document_query}")
            print(f"{colored('Q', 'green')}: {query}")
            answer, _ = await cybrex.chat_document(
                document_query,
                query,
                n_chunks,
                minimum_score=minimum_score,
            )
            print(f"{colored('A', 'green')}: {answer}")

    async def chat_sci(
        self,
        query: str,
        n_chunks: int = 5,
        n_documents: int = 10,
        minimum_score: float = 0.5,
    ):
        """
        Ask a general questions

        :param query: text query to the document
        :param n_chunks: the number of chunks to extract
            more means more tokens to use and more precision in answer
        :param n_documents: the number of chunks to extract
            more means more tokens to use and more precision in answer
        """
        async with self.cybrex as cybrex:
            print(f"{colored('Q', 'green')}: {query}")
            answer, chunks = await cybrex.chat_science(
                query=query,
                n_chunks=n_chunks,
                n_documents=n_documents,
                minimum_score=minimum_score,
            )
            answer = re.sub(r'\(DOI:\s*([^)]+)\)', r'(https://doi.org/\g<1>)', answer)
            references = []
            visited = set()
            for chunk in chunks:
                index_alias, field, value = chunk['document_id'].split(':', 2)
                document_id = f'{field}:{value}'
                if document_id in visited:
                    continue
                visited.add(document_id)
                references.append(f'{document_id}: {chunk["title"]}')
            references = '\n'.join(references)
            print(f"{colored('A', 'green')}: {answer}")
            print(f"{colored('References', 'green')}:\n{textwrap.indent(references, ' - ')}")

    async def sum_doc(self, document_query: str):
        """
        Summarization of the document

        :param document_query: query that returns unique document
        """
        async with self.cybrex as cybrex:
            print(f"{colored('Document', 'green')}: {document_query}")
            answer, _ = await cybrex.summarize_document(document_query)
            print(f"{colored('Summarization', 'green')}: {answer}")

    async def semantic_search(
        self,
        query: str,
        n_chunks: int = 5,
        n_documents: int = 10,
        minimum_score: float = 0.5,
    ):
        """
        Search related to query text chunks among `n` documents

        :param query: query to STC
        :param n_chunks: number of chunks to return
        :param n_documents: the number of documents to extract from STC
        :param minimum_score:
        """
        async with self.cybrex as cybrex:
            print(f"{colored('Q', 'green')}: {query}")
            chunks = await cybrex.semantic_search(
                query=query,
                n_chunks=n_chunks,
                n_documents=n_documents,
                minimum_score=minimum_score
            )
            references = []
            for chunk in chunks:
                index_alias, field, value = chunk['document_id'].split(':', 2)
                document_id = f'{field}:{value}'
                references.append(f' - {document_id}: {chunk["title"]}\n   {chunk["text"]}')
            references = '\n'.join(references)
            print(f"{colored('References', 'green')}:\n{references}")


def cybrex_cli(debug: bool = False):
    """
    :param debug: add debugging output
    :return:
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO if debug else logging.ERROR)
    cybrex = CybrexCli()
    return {
        'add-all-chunks': cybrex.add_all_documents,
        'chat-doc': cybrex.chat_doc,
        'chat-sci': cybrex.chat_sci,
        'export-chunks': cybrex.export_chunks,
        'import-chunks': cybrex.cybrex.import_chunks,
        'semantic-search': cybrex.semantic_search,
        'sum-doc': cybrex.sum_doc,
        'write-config': cybrex.cybrex.ensure_config,
    }


def main():
    fire.Fire(cybrex_cli, name='cybrex')


if __name__ == '__main__':
    main()



from typing import (
    List,
    Optional,
)

import orjson
from stc_geck.client import StcGeck

from .base import (
    BaseDataSource,
    SourceDocument,
)


class GeckDataSource(BaseDataSource):
    def __init__(self, geck: StcGeck):
        self.geck = geck

    async def query_documents(
        self,
        query: str,
        limit: int = 5,
        sources: Optional[List[str]] = None
    ) -> List[SourceDocument]:
        queries = self.geck.get_query_processor().process(query, limit=limit, index_aliases=sources)
        response = await self.geck.get_summa_client().search(queries)
        source_documents = []
        for scored_document in response.collector_outputs[0].documents.scored_documents:
            document = orjson.loads(scored_document.document)
            match scored_document.index_alias:
                case 'nexus_science':
                    document_id = f'nexus_science:doi:{document["doi"]}'
                case 'nexus_free':
                    if 'id' in document:
                        if 'isbns' in document['id']:
                            document_id = f'nexus_free:id.isbns:{document["id"]["isbns"][0]}'
                        elif 'internal_iso' in document['id']:
                            document_id = f'nexus_free:id.internal_iso:{document["id"]["internal_iso"]}'
                        else:
                            continue
                    else:
                        continue
                case _:
                    raise RuntimeError("Unsupported table")
            source_documents.append(SourceDocument(
                document=document,
                document_id=document_id,
            ))
        return source_documents



import logging


class LLMManager:
    def __init__(self, llm, prompter, config, max_prompt_chars, tokenizer=None):
        self.llm = llm
        self.prompter = prompter
        self.config = config
        self.max_prompt_chars = max_prompt_chars
        self.tokenizer = tokenizer

    @property
    def context_length(self):
        return self.config['context_length']

    def process(self, prompt):
        logging.getLogger('statbox').info({'action': 'process', 'mode': 'llm', 'prompt': prompt})
        if self.tokenizer:
            input_ids = self.tokenizer(prompt, return_tensors="pt")["input_ids"]
            outputs = self.llm.generate(
                input_ids,
                max_new_tokens=self.config.get('max_new_tokens'),
                temperature=self.config.get('temperature', 1.0),
            )
            return self.tokenizer.batch_decode(outputs[:, input_ids.shape[1]:])[0].replace('</s>', '')
        else:
            return self.llm(prompt)
